from pyslang import *


def test_numerics():
    assert SVInt(-3) == -3
    assert 99887766554433221100 == SVInt(99887766554433221100)
    assert clog2(99887766554433221100) == 67
    assert int(logic_t(0) | logic_t(1)) == 1
    assert int(SVInt("32'd1234")) == 1234

    cv = ConstantValue(3.14159)
    assert cv.value == 3.14159
    assert cv.isTrue()
    assert cv.convertToInt().value == 3

    r = ConstantRange(7, 1)
    assert str(r.reverse()) == "[1:7]"


def test_bag():
    lo = LexerOptions()
    po = ParserOptions()
    po.maxRecursionDepth = 4
    b = Bag([lo, po])
    assert b.parserOptions.maxRecursionDepth == 4
    assert b.compilationOptions is None


testFile = """
module m(input i, output o);
    assign o = (~i + 32'd1234);
endmodule
"""


def test_version():
    assert VersionInfo.getMajor() == 2


def test_sourcemanager():
    sm = SourceManager()
    buf = sm.assignText("coolfile.sv", testFile)

    filename = sm.getFileName(SourceLocation(buf.id, 0))
    assert filename == "coolfile.sv"


def test_syntax():
    tree = SyntaxTree.fromText(testFile)
    assert len(tree.diagnostics) == 0

    m = tree.root
    assert m.kind == SyntaxKind.ModuleDeclaration
    assert m.header.name.value == "m"

    a = m.members[0]
    assert a.kind == SyntaxKind.ContinuousAssign
    assert a.assignments[0].kind == SyntaxKind.AssignmentExpression

    zeroConst = a.assignments[0].right.expression.right
    assert zeroConst.kind == SyntaxKind.IntegerVectorExpression
    assert zeroConst.value.value == 1234


def test_compilation():
    tree = SyntaxTree.fromText(testFile)
    comp = Compilation()
    comp.addSyntaxTree(tree)

    diags = comp.getAllDiagnostics()
    assert len(diags) == 1
    assert diags[0].code == Diags.WidthTruncate

    report = DiagnosticEngine.reportAll(comp.sourceManager, diags)
    assert (
        ("\n" + report)
        == """
source:3:14: warning: implicit conversion truncates from 32 to 1 bits [-Wwidth-trunc]
    assign o = (~i + 32'd1234);
             ^  ~~~~~~~~~~~~~
"""
    )


def test_scriptsession():
    session = ScriptSession()
    session.eval('integer arr[string] = \'{"Hello":4, "World":8, default:-1};')
    session.eval(
        """
function int func(int i, integer arr[string]);
    case (i) inside
        [5'd1:5'd2]: return 1;
        arr: return 2;
    endcase
    return 3;
endfunction
"""
    )

    assert session.eval("func(4, arr)").value == 2
    assert len(session.getDiagnostics()) == 0



from pyslang import *

testfile = """
//! Module description
//! ***this is code*** sample
//! | Tables   |     Are      |  Cool |
//! |----------|:------------:|------:|
//! | col 1 is | left-aligned | $1600 |

module gray_counter (
    out    , // counter out
    clk    , //! clock
    clk1   , //! clock sample
    rst      //! **active high reset**
    );

    input clk, clk1, rst;
    output [7:0] out;
    wire [7:0] out;
    reg [7:0] count;

endmodule
"""


def test_extractcomments():
    tree = SyntaxTree.fromText(testfile)
    assert tree.root.kind == SyntaxKind.ModuleDeclaration

    moduleComments = []
    for t in tree.root.getFirstToken().trivia:
        if t.kind == TriviaKind.LineComment:
            comment = t.getRawText()
            if comment.startswith("//!"):
                moduleComments.append(comment[3:].strip())

    portComments = {}
    portList = tree.root.header.ports
    lastPort = None

    def getLeadingComments(tok):
        if lastPort is not None:
            for t in tok.trivia:
                if t.kind == TriviaKind.LineComment:
                    comment = t.getRawText()
                    if comment.startswith("//!"):
                        portComments[lastPort].append(comment[3:].strip())
                elif t.kind == TriviaKind.EndOfLine:
                    break

    if portList is not None:
        for port in portList.ports:
            if isinstance(port, ImplicitNonAnsiPortSyntax):
                tok = port.getFirstToken()
                getLeadingComments(tok)

                portName = tok.value
                portComments[portName] = []
                lastPort = portName

        getLeadingComments(portList.closeParen)

    assert len(moduleComments) == 5
    assert moduleComments[4] == "| col 1 is | left-aligned | $1600 |"

    for k, _ in portComments.copy().items():
        if len(portComments[k]) == 0:
            del portComments[k]

    assert len(portComments) == 3
    assert portComments["rst"][0] == "**active high reset**"



#!/usr/bin/env python
# This script generates C++ code for compiler diagnostics.
import argparse
import os
import shlex
import subprocess


def writefile(path, contents):
    try:
        with open(path, "r") as f:
            existing = f.read()
    except OSError:
        existing = ""

    if existing != contents:
        with open(path, "w") as f:
            f.write(contents)


def main():
    parser = argparse.ArgumentParser(description="Diagnostic source generator")
    parser.add_argument("--outDir", default=os.getcwd(), help="Output directory")
    parser.add_argument("--srcDir", help="Source directory to search for usages")
    parser.add_argument("--incDir", help="Include directory to search for usages")
    parser.add_argument("--docs", action="store_true")
    parser.add_argument("--slangBin")
    args = parser.parse_args()

    ourdir = os.path.dirname(os.path.realpath(__file__))
    inf = open(os.path.join(ourdir, "diagnostics.txt"))

    headerdir = os.path.join(args.outDir, "slang", "diagnostics")
    try:
        os.makedirs(headerdir)
    except OSError:
        pass

    diags = {}
    groups = []
    diaglist = []
    subsystem = "General"
    curgroup = None

    def parsegroup(elems):
        nonlocal curgroup
        for e in elems:
            if e == "}":
                groups.append(curgroup)
                curgroup = None
                break
            curgroup[1].append(e)

    for line in [x.strip("\n") for x in inf]:
        if not line or line.startswith("//"):
            continue

        parts = shlex.split(line)
        if curgroup:
            parsegroup(parts)
        elif parts[0] == "subsystem":
            subsystem = parts[1]
            if subsystem not in diags:
                diags[subsystem] = []
        elif parts[0] == "group":
            curgroup = (parts[1], [])
            assert parts[2] == "="
            assert parts[3] == "{"
            parsegroup(parts[4:])
        else:
            sev = parts[0]
            if sev == "warning":
                diags[subsystem].append(("Warning", parts[2], parts[3], parts[1]))
                diaglist.append(parts[2])
            elif sev == "error":
                diags[subsystem].append(("Error", parts[1], parts[2], ""))
                diaglist.append(parts[1])
            elif sev == "note":
                diags[subsystem].append(("Note", parts[1], parts[2], ""))
                diaglist.append(parts[1])
            else:
                raise Exception("Invalid entry: {}".format(line))

    if args.docs:
        createdocs(
            args.outDir,
            os.path.join(ourdir, "warning_docs.txt"),
            args.slangBin,
            diags,
            groups,
        )
    else:
        for k, v in sorted(diags.items()):
            createheader(os.path.join(headerdir, k + "Diags.h"), k, v)

        createsource(os.path.join(args.outDir, "DiagCode.cpp"), diags, groups)
        createallheader(os.path.join(headerdir, "AllDiags.h"), diags)

        doCheck = False
        if doCheck:
            diaglist = checkDiags(args.srcDir, diaglist)
            diaglist = checkDiags(args.incDir, diaglist)
            reportUnused(diaglist)


def createheader(path, subsys, diags):
    output = """//------------------------------------------------------------------------------
//! @file {}Diags.h
//! @brief Generated diagnostic enums for the {} subsystem
//
// File is under the MIT license; see LICENSE for details
//------------------------------------------------------------------------------
#pragma once

#include "slang/diagnostics/Diagnostics.h"

namespace slang::diag {{

""".format(
        subsys, subsys
    )

    index = 0
    for d in sorted(diags):
        output += "inline constexpr DiagCode {}(DiagSubsystem::{}, {});\n".format(
            d[1], subsys, index
        )
        index += 1

    output += """
}
"""
    writefile(path, output)


def createsource(path, diags, groups):
    output = """//------------------------------------------------------------------------------
// DiagCode.cpp
// Generated diagnostic helpers
//
// File is under the MIT license; see LICENSE for details
//------------------------------------------------------------------------------
#include "slang/diagnostics/AllDiags.h"

#include <ostream>

#include "slang/util/Hash.h"

namespace slang {

static const flat_hash_map<DiagCode, std::tuple<string_view, string_view, DiagnosticSeverity, string_view>> data = {
"""

    for _, v in sorted(diags.items()):
        for d in sorted(v):
            output += '    {{diag::{}, std::make_tuple("{}"sv, "{}"sv, DiagnosticSeverity::{}, "{}"sv)}},\n'.format(
                d[1], d[1], d[2], d[0], d[3]
            )

    output += """};

static const flat_hash_map<string_view, std::vector<DiagCode>> optionMap = {
"""

    optionMap = {}
    for _, v in sorted(diags.items()):
        for d in sorted(v):
            name = d[3]
            if not name:
                continue

            if name in optionMap:
                optionMap[name].append(d[1])
            else:
                optionMap[name] = [d[1]]

    for key in sorted(optionMap):
        vals = optionMap[key]
        valstr = ", ".join(["diag::{}".format(v) for v in vals])
        output += '    {{"{}"sv, {{ {} }}}},\n'.format(key, valstr)

    output += """};

static const flat_hash_map<string_view, DiagGroup> groupMap = {
"""

    for g in sorted(groups):
        elems = []
        for e in sorted(g[1]):
            elems.extend(optionMap[e])

        elems = ", ".join("diag::{}".format(e) for e in elems)
        output += '    {{"{}"sv, DiagGroup("{}", {{ {} }})}},\n'.format(
            g[0], g[0], elems
        )

    output += """};

std::ostream& operator<<(std::ostream& os, DiagCode code) {
    os << toString(code);
    return os;
}

string_view toString(DiagCode code) {
    if (auto it = data.find(code); it != data.end())
        return std::get<0>(it->second);
    return "<user-diag>"sv;
}

string_view getDefaultMessage(DiagCode code) {
    if (auto it = data.find(code); it != data.end())
        return std::get<1>(it->second);
    return ""sv;
}

DiagnosticSeverity getDefaultSeverity(DiagCode code) {
    if (auto it = data.find(code); it != data.end())
        return std::get<2>(it->second);
    return DiagnosticSeverity::Ignored;
}

string_view getDefaultOptionName(DiagCode code) {
    if (auto it = data.find(code); it != data.end())
        return std::get<3>(it->second);
    return ""sv;
}

span<const DiagCode> findDiagsFromOptionName(string_view name) {
    if (auto it = optionMap.find(name); it != optionMap.end())
        return it->second;
    return {};
}

const DiagGroup* findDefaultDiagGroup(string_view name) {
    if (auto it = groupMap.find(name); it != groupMap.end())
        return &it->second;
    return nullptr;
}

static const DiagCode AllGeneratedCodes[] = {
"""

    for _, v in sorted(diags.items()):
        for d in sorted(v):
            output += "    diag::{},\n".format(d[1])

    output += """};

decltype(DiagCode::KnownCodes) DiagCode::KnownCodes = AllGeneratedCodes;

}"""

    writefile(path, output)


def createallheader(path, diags):
    output = """//------------------------------------------------------------------------------
//! @file AllDiags.h
//! @brief Combined header that includes all subsystem-specific diagnostic headers
//
// File is under the MIT license; see LICENSE for details
//------------------------------------------------------------------------------
#pragma once

"""

    for k in sorted(diags.keys()):
        output += '#include "slang/diagnostics/{}Diags.h"\n'.format(k)

    output += "\n"
    writefile(path, output)


def createdocs(outDir, inpath, slangBin, diags, groups):
    inf = open(inpath)
    curropt = None
    inexample = False
    exampleMap = {}

    for line in [x.strip("\n") for x in inf]:
        if not inexample:
            line = line.strip()
            if not line or line.startswith("//"):
                continue

        if inexample:
            if line.startswith("```"):
                inexample = False
            else:
                if curropt[2]:
                    curropt[2] += "\n"
                curropt[2] += line
        elif line.startswith("-W"):
            if curropt:
                exampleMap[curropt[0]] = curropt
            curropt = [line[2:], "", "", ""]
        elif line.startswith("```"):
            inexample = True
        else:
            if curropt[1]:
                curropt[1] += " "
            curropt[1] += line

    if curropt:
        exampleMap[curropt[0]] = curropt

    for k, v in exampleMap.items():
        if not v[2]:
            continue

        testPath = os.path.join(outDir, "test.sv")
        with open(testPath, "w") as outf:
            outf.write(v[2])

        encoding = "utf-16-le" if os.name == "nt" else "utf-8"
        args = [
            slangBin,
            "--quiet",
            "-Wnone",
            "-W" + k,
            "--color-diagnostics",
            testPath,
        ]

        result = subprocess.run(
            args, encoding=encoding, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
        v[3] = result.stdout
        if k not in v[3]:
            raise Exception("Test for -W{} is not correct".format(k))

    output = """/** @page warning-ref Warning Reference
@brief Reference information about all supported warnings

@tableofcontents

@section warnings Warnings

"""

    groupMap = {}
    warnlist = []
    for g in groups:
        if g[0] != "default":
            warnlist.append(g)

        for e in g[1]:
            if e in groupMap:
                groupMap[e].add(g[0])
            else:
                groupMap[e] = set([g[0]])

    for _, v in diags.items():
        for d in v:
            if not d[3]:
                continue
            warnlist.append(d)

    warnlist.sort(key=lambda d: d[3] if len(d) > 3 else d[0])

    lastOpt = ""
    for d in warnlist:
        if len(d) > 3:
            opt = d[3]
            if opt == lastOpt:
                continue

            if opt not in exampleMap:
                raise Exception("No documentation for -W{}".format(opt))

            details = exampleMap[opt]
            desc = details[1]
            example = details[2]
            results = details[3]

            if desc == "<ignored>":
                continue

            if lastOpt != "":
                output += "\n@n\n"
            output += "@subsection {} -W{}\n".format(opt, opt)

            output += desc
            output += " @n @n\n"

            if opt in groupMap:
                groups = groupMap[opt]
                if "default" in groups:
                    output += "This diagnostic is enabled by default. @n @n\n"

            if example:
                assert results
                output += "@b Example: \n\n"
                output += "@code{.sv}\n"
                output += example + "\n"
                output += "@endcode\n\n"
                output += "produces:\n\n"
                output += "@code{.ansi}\n"
                output += results
                output += "@endcode\n"

            lastOpt = opt
        else:
            opt = d[0]
            lastOpt = opt
            elemlist = ", ".join("@ref {}".format(s) for s in d[1])

            output += "\n@n\n@subsection {} -W{}\n".format(opt, opt)
            output += "Controls {}.\n@n\n".format(elemlist)

    output += "\n*/"
    writefile(os.path.join(outDir, "warnings.dox"), output)


def checkDiags(path, diags):
    import glob

    for ext in ("cpp", "h"):
        for file in glob.glob(path + "/**/*." + ext, recursive=True):
            with open(file, "r") as f:
                text = f.read()
                diags = [d for d in diags if not ("::" + d) in text]
    return diags


def reportUnused(diags):
    for d in diags:
        print("warning: '{}' is unused".format(d))


if __name__ == "__main__":
    main()



from subprocess import check_output

from setuptools import find_packages
from skbuild import setup

command = "git describe --tags --long --dirty"
fmt = "{tag}+{commitcount}.{gitsha}"

git_version = check_output(command.split()).decode("utf-8").strip()
if git_version.startswith("v"):
    git_version = git_version[1:]

parts = git_version.split("-")
assert len(parts) in (3, 4)
dirty = len(parts) == 4
tag, count, sha = parts[:3]
if count == "0" and not dirty:
    version = tag
else:
    version = fmt.format(tag=tag, commitcount=count, gitsha=sha)
    if dirty:
        version = version + ".dirty"

setup(
    version=version,
    packages=[""],
    package_dir={"": "source"},
    cmake_source_dir="../",
    cmake_install_dir="source",
    cmake_install_target="pyslang-install-pylib",
    cmake_args=[
        "-DSLANG_INCLUDE_TESTS=OFF",
        "-DSLANG_INCLUDE_TOOLS=OFF",
        "-DSLANG_INCLUDE_PYLIB=ON",
    ],
    include_package_data=True,
    extras_require={"test": ["pytest"]},
)



#!/usr/bin/env python
# Takes an input text file containing the SystemVerilog language grammar
# and converts it to nice linked markdown.

import os
import re

ourdir = os.path.dirname(os.path.realpath(__file__))

inf = open(os.path.join(ourdir, "grammar.txt"))
outf = open(os.path.join(ourdir, "../docs/grammar.md"), "w")

outf.write("# SystemVerilog\n")


def entry(line):
    match = re.match(r"(\w+) ::=", line)
    if match:
        s = match.group(1)
        outf.write('<a name="{}"></a>{} ::='.format(s, s.replace("_", "\\_")))
        line = line[len(match.group(0)) :]
    else:
        outf.write("&nbsp;&nbsp;&nbsp;&nbsp;")

    saved_match = None
    match = re.search(r"(\w+) ::=", line)
    if match:
        saved_match = line[match.start() :]
        line = line[: match.start()]

    def replacer(m):
        s = m.group(1)
        t = s
        for c in ["*", "-", "|", "[", "{", "_"]:
            t = t.replace(c, "\\" + c)
        if s not in ["|", "[", "]", "{", "}"]:
            return "`{}`".format(s)
        return t

    line = line.replace("{,", "{ ,")
    line = re.sub(r"([^\w\s]+)", replacer, line)
    line = re.sub(r"(\w+)", r"[\1](#\1)", line)
    outf.write(line + "  \n")

    if saved_match:
        entry(saved_match)


for line in inf:
    line = line.strip()
    if not line:
        continue

    if line.startswith("A."):
        count = line.split(" ")[0].count(".")
        if count == 1:
            outf.write("## ")
        elif count == 2:
            outf.write("### ")
        elif count == 3:
            outf.write("#### ")
        else:
            raise Exception("wut")
        outf.write(line + "\n")
    else:
        entry(line)



#!/usr/bin/env python
# This script generates C++ source for parse tree syntax nodes from a data file.
import argparse
import os


class TypeInfo:
    def __init__(
        self,
        processedMembers,
        members,
        pointerMembers,
        optionalMembers,
        final,
        constructorArgs,
        argNames,
        base,
        combinedMembers,
        notNullMembers,
        kindValue,
        initializers,
        baseInitializers,
    ):
        self.processedMembers = processedMembers
        self.members = members
        self.pointerMembers = pointerMembers
        self.optionalMembers = optionalMembers
        self.final = final
        self.constructorArgs = constructorArgs
        self.argNames = argNames
        self.base = base
        self.combinedMembers = combinedMembers
        self.notNullMembers = notNullMembers
        self.kindValue = kindValue
        self.initializers = initializers
        self.baseInitializers = baseInitializers


def main():
    parser = argparse.ArgumentParser(description="Diagnostic source generator")
    parser.add_argument("--dir", default=os.getcwd(), help="Output directory")
    parser.add_argument("--python-bindings", action="store_true")
    args = parser.parse_args()

    ourdir = os.path.dirname(os.path.realpath(__file__))
    alltypes, kindmap = loadalltypes(ourdir)

    if args.python_bindings:
        generatePyBindings(args.dir, alltypes)
    else:
        generateSyntax(args.dir, alltypes, kindmap)
        generateTokenKinds(ourdir, args.dir)


def loadalltypes(ourdir):
    inf = open(os.path.join(ourdir, "syntax.txt"))

    currtype = None
    currkind = None
    currtype_name = None
    tags = None
    alltypes = {}
    kindmap = {}

    alltypes["SyntaxNode"] = TypeInfo(
        None, None, None, None, "", None, None, None, [], None, "", None, None
    )

    for line in [x.strip("\n") for x in inf]:
        if line.startswith("//"):
            continue
        elif not line or (currtype is not None and line == "empty"):
            if currtype is not None:
                createtype(currtype_name, tags, currtype, alltypes, kindmap)
            currtype = None
            currkind = None
        elif currtype is not None:
            p = line.split(" ")
            if len(p) != 2:
                raise Exception("Two elements per member please.")
            currtype.append(p)
        elif currkind is not None:
            for k in line.split(" "):
                if k in kindmap:
                    raise Exception("More than one kind map for {}".format(k))
                kindmap[k] = currkind
        elif line.startswith("kindmap<"):
            currkind = line[8 : line.index(">")] + "Syntax"
        else:
            p = line.split(" ")
            currtype_name = p[0] + "Syntax"
            tags = p[1:] if len(p) > 1 else None
            currtype = []

    if currtype:
        createtype(currtype_name, tags, currtype, alltypes, kindmap)

    return (alltypes, kindmap)


def createtype(name, tags, members, alltypes, kindmap):
    tagdict = {}
    if tags:
        for t in tags:
            p = t.split("=")
            tagdict[p[0]] = p[1]

    base = tagdict["base"] + "Syntax" if "base" in tagdict else "SyntaxNode"

    pointerMembers = set()
    optionalMembers = set()
    notNullMembers = set()
    processedMembers = []
    baseInitializers = ""
    combined = members
    if base != "SyntaxNode":
        processedMembers.extend(alltypes[base].processedMembers)
        pointerMembers = pointerMembers.union(alltypes[base].pointerMembers)
        optionalMembers = optionalMembers.union(alltypes[base].optionalMembers)
        notNullMembers = notNullMembers.union(alltypes[base].notNullMembers)
        baseInitializers = ", ".join([x[1] for x in alltypes[base].members])
        if baseInitializers:
            baseInitializers = ", " + baseInitializers
        combined = alltypes[base].members + members

    for m in members:
        if m[0] == "token":
            m[0] = "Token"
            typename = m[0]
        elif m[0] == "tokenlist":
            m[0] = "TokenList"
            typename = m[0]
            pointerMembers.add(m[1])
        elif m[0].startswith("list<"):
            last = m[0][5 : m[0].index(">")]
            if not last.endswith("SyntaxNode"):
                last += "Syntax"

            m[0] = "SyntaxList<" + last + ">"
            typename = m[0]
            pointerMembers.add(m[1])
        elif m[0].startswith("separated_list<"):
            last = m[0][15 : m[0].index(">")]
            if not last.endswith("SyntaxNode"):
                last += "Syntax"

            m[0] = "SeparatedSyntaxList<" + last + ">"
            typename = m[0]
            pointerMembers.add(m[1])
        else:
            optional = False
            if m[0].endswith("?"):
                optional = True
                m[0] = m[0][:-1]

            if m[0] != "SyntaxNode":
                m[0] += "Syntax"

            if m[0] not in alltypes:
                raise Exception("Unknown type '{}'".format(m[0]))

            typename = m[0]
            if optional:
                m[0] += "*"
                optionalMembers.add(m[1])
            else:
                m[0] += "&"
                notNullMembers.add(m[1])

        m.append(typename)
        processedMembers.append("{} {}".format(m[0], m[1]))
        if m[1] in notNullMembers:
            m[0] = "not_null<{}*>".format(typename)

    final = " final"
    if "final" in tagdict and tagdict["final"] == "false":
        final = ""

    kindArg = "SyntaxKind kind"
    kindValue = "kind"
    argNames = []

    if not final or tagdict.get("multiKind") == "true":
        argNames.append("kind")
    else:
        k = name
        if k.endswith("Syntax"):
            k = k[:-6]

        if k in kindmap:
            raise Exception("More than one kind map for {}".format(k))
        kindmap[k] = name
        kindArg = ""
        kindValue = "SyntaxKind::" + k

    if kindArg and processedMembers:
        kindArg += ", "

    initializers = ", ".join(
        [
            "{0}({1}{0})".format(x[1], "&" if x[1] in notNullMembers else "")
            for x in members
        ]
    )
    if initializers:
        initializers = ", " + initializers

    argMembers = []
    for m in processedMembers:
        space = m.index(" ")
        argNames.append(m[space + 1 :])

        if (
            m.startswith("SyntaxList<")
            or m.startswith("SeparatedSyntaxList<")
            or m.startswith("TokenList")
        ):
            argMembers.append("const {}&{}".format(m[:space], m[space:]))
        else:
            argMembers.append(m)

    if final and not argMembers:
        raise Exception("{} has no members".format(name))

    constructorArgs = "{}{}".format(kindArg, ", ".join(argMembers))
    alltypes[name] = TypeInfo(
        processedMembers,
        members,
        pointerMembers,
        optionalMembers,
        final,
        constructorArgs,
        argNames,
        base,
        combined,
        notNullMembers,
        kindValue,
        initializers,
        baseInitializers,
    )


def generateSyntax(builddir, alltypes, kindmap):
    # Make sure the output directory exists.
    headerdir = os.path.join(builddir, "slang", "syntax")
    try:
        os.makedirs(headerdir)
    except OSError:
        pass

    # Start the header file.
    outf = open(os.path.join(headerdir, "AllSyntax.h"), "w")
    outf.write(
        """//------------------------------------------------------------------------------
//! @file AllSyntax.h
//! @brief All generated syntax node data structures
//
// File is under the MIT license; see LICENSE for details
//------------------------------------------------------------------------------
#pragma once

#include "slang/parsing/Token.h"
#include "slang/syntax/SyntaxNode.h"
#include "slang/util/BumpAllocator.h"

// This file contains all parse tree syntax nodes.
// It is auto-generated by the syntax_gen.py script under the scripts/ directory.

namespace slang {

"""
    )

    # Write all type definitions.
    for name, currtype in alltypes.items():
        if name == "SyntaxNode":
            continue

        outf.write("struct {} : public {} {{\n".format(name, currtype.base))

        for m in currtype.members:
            outf.write("    {} {};\n".format(m[0], m[1]))

        outf.write("\n")
        outf.write("    {}({}) :\n".format(name, currtype.constructorArgs))
        outf.write(
            "        {}({}{}){} {{\n".format(
                currtype.base,
                currtype.kindValue,
                currtype.baseInitializers,
                currtype.initializers,
            )
        )

        # Write constructor body.
        for m in currtype.members:
            if m[0] == "Token":
                continue
            if m[1] in currtype.pointerMembers:
                outf.write("        this->{}.parent = this;\n".format(m[1]))
                if m[0].startswith("SyntaxList<") or m[0].startswith(
                    "SeparatedSyntaxList<"
                ):
                    outf.write("        for (auto child : this->{})\n".format(m[1]))
                    outf.write("            child->parent = this;\n")
            elif m[1] in currtype.optionalMembers:
                outf.write(
                    "        if (this->{0}) this->{0}->parent = this;\n".format(m[1])
                )
            else:
                outf.write("        this->{}->parent = this;\n".format(m[1]))

        outf.write("    }\n\n")

        # Copy constructor (defaulted).
        outf.write("    explicit {}(const {}&) = default;\n\n".format(name, name))

        if not currtype.members and currtype.final == "":
            outf.write("    static bool isKind(SyntaxKind kind);\n")
        else:
            outf.write("    static bool isKind(SyntaxKind kind);\n\n")

            outf.write("    TokenOrSyntax getChild(size_t index);\n")
            outf.write("    ConstTokenOrSyntax getChild(size_t index) const;\n")
            outf.write("    void setChild(size_t index, TokenOrSyntax child);\n\n")
            outf.write("    {}* clone(BumpAllocator& alloc) const;\n".format(name))

        outf.write("};\n\n")

    # Start the source file.
    cppf = open(os.path.join(builddir, "AllSyntax.cpp"), "w")
    cppf.write(
        """//------------------------------------------------------------------------------
// AllSyntax.cpp
// All generated syntax node data structures
//
// File is under the MIT license; see LICENSE for details
//------------------------------------------------------------------------------
#include "slang/syntax/AllSyntax.h"

// This file contains all parse tree syntax node generated definitions.
// It is auto-generated by the syntax_gen.py script under the scripts/ directory.

namespace slang {

size_t SyntaxNode::getChildCount() const {
    switch (kind) {
        case SyntaxKind::Unknown: return 0;
        case SyntaxKind::SyntaxList:
        case SyntaxKind::TokenList:
        case SyntaxKind::SeparatedList:
            return ((const SyntaxListBase*)this)->getChildCount();
"""
    )

    for k, v in sorted(kindmap.items()):
        count = len(alltypes[v].combinedMembers)
        cppf.write("        case SyntaxKind::{}: return {};\n".format(k, count))

    cppf.write("    }\n")
    cppf.write("    THROW_UNREACHABLE;\n")
    cppf.write("}\n\n")

    # Build a reverse mapping from class types to their syntax kinds.
    reverseKindmap = {}
    for k, v in kindmap.items():
        if v in reverseKindmap:
            reverseKindmap[v].append(k)
        else:
            reverseKindmap[v] = [k]

    # Continue building up the reverse map by traversing base class links.
    for k, v in alltypes.items():
        if not v.final:
            continue

        while v.base != "SyntaxNode":
            kinds = reverseKindmap[k]
            if v.base in reverseKindmap:
                reverseKindmap[v.base].extend(kinds)
            else:
                reverseKindmap[v.base] = kinds[:]
            k = v.base
            v = alltypes[k]

    # Write out isKind static methods for each derived type.
    for k, v in sorted(alltypes.items()):
        if v.base is None:
            continue

        cppf.write("bool {}::isKind(SyntaxKind kind) {{\n".format(k))
        kinds = set(reverseKindmap[k])
        if len(kinds) == 1:
            cppf.write("    return kind == SyntaxKind::{};\n".format(list(kinds)[0]))
        else:
            cppf.write("    switch (kind) {\n")
            for kind in sorted(kinds):
                cppf.write("        case SyntaxKind::{}:\n".format(kind))
            cppf.write("            return true;\n")
            cppf.write("        default:\n")
            cppf.write("            return false;\n")
            cppf.write("    }\n")

        cppf.write("}\n\n")

        if v.members or v.final != "":
            for returnType in ("TokenOrSyntax", "ConstTokenOrSyntax"):
                cppf.write(
                    "{} {}::getChild(size_t index){} {{\n".format(
                        returnType, k, "" if returnType == "TokenOrSyntax" else " const"
                    )
                )

                if v.combinedMembers:
                    cppf.write("    switch (index) {\n")

                    index = 0
                    for m in v.combinedMembers:
                        addr = "&" if m[1] in v.pointerMembers else ""
                        get = ".get()" if m[1] in v.notNullMembers else ""
                        cppf.write(
                            "        case {}: return {}{}{};\n".format(
                                index, addr, m[1], get
                            )
                        )
                        index += 1

                    cppf.write("        default: return nullptr;\n")
                    cppf.write("    }\n")
                else:
                    cppf.write("    (void)index;\n")
                    cppf.write("    return nullptr;\n")

                cppf.write("}\n\n")

            cppf.write(
                "void {}::setChild(size_t index, TokenOrSyntax child) {{\n".format(k)
            )
            if v.combinedMembers:
                cppf.write("    switch (index) {\n")

                index = 0
                for m in v.combinedMembers:
                    cppf.write("        case {}: ".format(index))
                    index += 1

                    if m[0] == "Token":
                        cppf.write("{} = child.token(); return;\n".format(m[1]))
                    elif m[1] in v.pointerMembers:
                        cppf.write(
                            "{} = child.node()->as<{}>(); return;\n".format(m[1], m[2])
                        )
                    else:
                        cppf.write(
                            "{} = &child.node()->as<{}>(); return;\n".format(m[1], m[2])
                        )

                cppf.write("        default: THROW_UNREACHABLE;\n")
                cppf.write("    }\n")
            else:
                cppf.write("    (void)index;\n")
                cppf.write("    (void)child;\n")

            cppf.write("}\n\n")

            cppf.write("{0}* {0}::clone(BumpAllocator& alloc) const {{\n".format(k))
            cppf.write("    return alloc.emplace<{}>(*this);\n".format(k))
            cppf.write("}\n\n")

    # Write out syntax factory methods.
    outf.write("class SyntaxFactory {\n")
    outf.write("public:\n")
    outf.write("    explicit SyntaxFactory(BumpAllocator& alloc) : alloc(alloc) {}\n")
    outf.write("\n")

    for k, v in sorted(alltypes.items()):
        if not v.final:
            continue

        methodName = k
        if methodName.endswith("Syntax"):
            methodName = methodName[:-6]
        methodName = methodName[:1].lower() + methodName[1:]
        outf.write("    {}& {}({});\n".format(k, methodName, v.constructorArgs))

        argNames = ", ".join(v.argNames)
        cppf.write(
            "{}& SyntaxFactory::{}({}) {{\n".format(k, methodName, v.constructorArgs)
        )
        cppf.write("    return *alloc.emplace<{}>({});\n".format(k, argNames))
        cppf.write("}\n\n")

    # Write out toString methods for SyntaxKind enum.
    cppf.write(
        """
std::ostream& operator<<(std::ostream& os, SyntaxKind kind) {
    os << toString(kind);
    return os;
}

string_view toString(SyntaxKind kind) {
    switch (kind) {
        case SyntaxKind::Unknown: return "Unknown";
        case SyntaxKind::SyntaxList: return "SyntaxList";
        case SyntaxKind::TokenList: return "TokenList";
        case SyntaxKind::SeparatedList: return "SeparatedList";
"""
    )

    for k, _ in sorted(kindmap.items()):
        cppf.write('        case SyntaxKind::{}: return "{}";\n'.format(k, k))

    cppf.write(
        """        default: return "";
    }
}

"""
    )

    # Write out traits member list for SyntaxKind enum.
    cppf.write("decltype(SyntaxKind_traits::values) SyntaxKind_traits::values = {\n")
    cppf.write(
        """    SyntaxKind::Unknown,
    SyntaxKind::SyntaxList,
    SyntaxKind::TokenList,
    SyntaxKind::SeparatedList,
"""
    )
    for k, _ in sorted(kindmap.items()):
        cppf.write("    SyntaxKind::{},\n".format(k))
    cppf.write(
        """};

const std::type_info* typeFromSyntaxKind(SyntaxKind kind) {
    switch (kind) {
        case SyntaxKind::Unknown: break;
        case SyntaxKind::SyntaxList:
        case SyntaxKind::TokenList:
        case SyntaxKind::SeparatedList:
            return &typeid(SyntaxNode);
"""
    )

    for k, v in sorted(kindmap.items()):
        cppf.write("        case SyntaxKind::{}: return &typeid({});\n".format(k, v))
    cppf.write(
        """    }
    return nullptr;
}

}
"""
    )

    outf.write("\n")
    outf.write("private:\n")
    outf.write("    BumpAllocator& alloc;\n")
    outf.write("};\n\n")

    # Write out a dispatch method to get from SyntaxKind to actual concrete type
    outf.write("namespace detail {\n\n")
    outf.write("template<typename TNode, typename TVisitor, typename... Args>\n")
    outf.write(
        "decltype(auto) visitSyntaxNode(TNode* node, TVisitor& visitor, Args&&... args) {\n"
    )
    outf.write("    static constexpr bool isConst = std::is_const_v<TNode>;")
    outf.write("    switch (node->kind) {\n")
    outf.write(
        "        case SyntaxKind::Unknown: return visitor.visitInvalid(*node, std::forward<Args>(args)...);\n"
    )
    outf.write("        case SyntaxKind::SyntaxList:\n")
    outf.write("        case SyntaxKind::TokenList:\n")
    outf.write("        case SyntaxKind::SeparatedList:\n")
    outf.write(
        "            return visitor.visit(*static_cast<std::conditional_t<isConst, const SyntaxListBase*, SyntaxListBase*>>(node), std::forward<Args>(args)...);\n"
    )

    for k, v in sorted(kindmap.items()):
        outf.write(
            "        case SyntaxKind::{}: return visitor.visit(*static_cast<".format(k)
        )
        outf.write(
            "std::conditional_t<isConst, const {0}*, {0}*>>(node), std::forward<Args>(args)...);\n".format(
                v
            )
        )
        alltypes.pop(v, None)

    outf.write("    }\n")
    outf.write("    THROW_UNREACHABLE;\n")
    outf.write("}\n\n")
    outf.write("}\n\n")

    outf.write("template<typename TVisitor, typename... Args>\n")
    outf.write(
        "decltype(auto) SyntaxNode::visit(TVisitor& visitor, Args&&... args) {\n"
    )
    outf.write(
        "    return detail::visitSyntaxNode(this, visitor, std::forward<Args>(args)...);\n"
    )
    outf.write("}\n\n")

    outf.write("template<typename TVisitor, typename... Args>\n")
    outf.write(
        "decltype(auto) SyntaxNode::visit(TVisitor& visitor, Args&&... args) const {\n"
    )
    outf.write(
        "    return detail::visitSyntaxNode(this, visitor, std::forward<Args>(args)...);\n"
    )
    outf.write("}\n\n")

    outf.write("}\n")

    # Do some checking to make sure all types have at least one kind assigned,
    # or has set final=false.  We already removed types from alltypes in the
    # loop above.
    for k, v in alltypes.items():
        if v.final:
            print("Type '{}' has no kinds assigned to it.".format(k))

    # Write out the SyntaxKind header file.
    outf = open(os.path.join(headerdir, "SyntaxKind.h"), "w")
    outf.write(
        """//------------------------------------------------------------------------------
//! @file SyntaxKind.h
//! @brief Generated SyntaxKind enum
//
// File is under the MIT license; see LICENSE for details
//------------------------------------------------------------------------------
#pragma once

#include "slang/util/Enum.h"

namespace std { class type_info; }

namespace slang {

enum class SyntaxKind {
    Unknown,
    SyntaxList,
    TokenList,
    SeparatedList,
"""
    )

    for k, _ in sorted(kindmap.items()):
        outf.write("    {},\n".format(k))

    outf.write(
        """}};

std::ostream& operator<<(std::ostream& os, SyntaxKind kind);
string_view toString(SyntaxKind kind);

class SyntaxKind_traits {{
public:
    static const std::array<SyntaxKind, {}> values;
}};

const std::type_info* typeFromSyntaxKind(SyntaxKind kind);

}}
""".format(
            len(kindmap.items()) + 4
        )
    )


def loadkinds(ourdir, filename):
    kinds = []
    inf = open(os.path.join(ourdir, filename))
    for line in [x.strip("\n") for x in inf]:
        line = line.strip()
        if not line:
            continue

        kinds.append(line)
    return kinds


def writekinddecl(outf, name, basetype, kinds):
    outf.write("enum class {} : {} {{\n".format(name, basetype))
    for k in kinds:
        outf.write("    {},\n".format(k))

    outf.write(
        """}};

std::ostream& operator<<(std::ostream& os, {} kind);
string_view toString({} kind);

class {}_traits {{
public:
    static const std::array<{}, {}> values;
}};

""".format(
            name, name, name, name, len(kinds)
        )
    )


def writekindimpls(outf, name, kinds):
    outf.write(
        """std::ostream& operator<<(std::ostream& os, {} kind) {{
    os << toString(kind);
    return os;
}}

string_view toString({} kind) {{
    switch (kind) {{
""".format(
            name, name
        )
    )

    for k in kinds:
        outf.write('        case {}::{}: return "{}";\n'.format(name, k, k))
    outf.write(
        """        default: return "";
    }
}

"""
    )

    outf.write(
        """decltype({}_traits::values) {}_traits::values = {{
""".format(
            name, name
        )
    )

    for k in kinds:
        outf.write("    {}::{},\n".format(name, k))
    outf.write(
        """};

"""
    )


def generateTokenKinds(ourdir, builddir):
    headerdir = os.path.join(builddir, "slang", "parsing")
    try:
        os.makedirs(headerdir)
    except OSError:
        pass

    triviakinds = loadkinds(ourdir, "triviakinds.txt")
    tokenkinds = loadkinds(ourdir, "tokenkinds.txt")

    outf = open(os.path.join(headerdir, "TokenKind.h"), "w")
    outf.write(
        """//------------------------------------------------------------------------------
//! @file TokenKind.h
//! @brief Generated TokenKind and TriviaKind enums
//
// File is under the MIT license; see LICENSE for details
//------------------------------------------------------------------------------
#pragma once

#include <ostream>

#include "slang/util/Util.h"

namespace slang {

"""
    )

    writekinddecl(outf, "TriviaKind", "uint8_t", triviakinds)
    writekinddecl(outf, "TokenKind", "uint16_t", tokenkinds)
    outf.write("}\n")

    outf = open(os.path.join(builddir, "TokenKind.cpp"), "w")
    outf.write(
        """//------------------------------------------------------------------------------
// TokenKind.cpp
// Generated TokenKind and TriviaKind enums
//
// File is under the MIT license; see LICENSE for details
//------------------------------------------------------------------------------
#include "slang/parsing/TokenKind.h"

namespace slang {

"""
    )

    writekindimpls(outf, "TriviaKind", triviakinds)
    writekindimpls(outf, "TokenKind", tokenkinds)
    outf.write("}\n")


def generatePyBindings(builddir, alltypes):
    outf = open(os.path.join(builddir, "PySyntaxBindings.cpp"), "w")
    outf.write(
        """//------------------------------------------------------------------------------
// PySyntaxBindings.cpp
// Generated Python bindings for syntax types
//
// File is under the MIT license; see LICENSE for details
//------------------------------------------------------------------------------
#include "pyslang.h"

#include "slang/syntax/AllSyntax.h"

void registerSyntaxNodes(py::module_& m) {
"""
    )

    for k, v in alltypes.items():
        if k == "SyntaxNode":
            continue

        outf.write('    py::class_<{}, {}>(m, "{}")'.format(k, v.base, k))
        for m in v.members:
            outf.write('\n        .def_readwrite("{}", &{}::{})'.format(m[1], k, m[1]))
        outf.write(";\n\n")

    outf.write("}\n")


if __name__ == "__main__":
    main()



from subprocess import check_output

from skbuild import setup


def get_git_version():
    command = "git describe --tags --long --dirty"
    fmt = "{tag}.{commitcount}"

    try:
        git_version = check_output(command.split(), cwd="slang").decode("utf-8").strip()
    except Exception:
        return None

    if git_version.startswith("v"):
        git_version = git_version[1:]

    parts = git_version.split("-")
    assert len(parts) in (3, 4)
    dirty = len(parts) == 4
    tag, count, sha = parts[:3]
    if count == "0" and not dirty:
        version = tag
    else:
        version = fmt.format(tag=tag, commitcount=count)
        if dirty:
            version = version + ".dirty"

    return version


def get_version():
    # Try to read existing release version file.
    try:
        f = open("RELEASE-VERSION", "r")
        fs_version = f.readlines()[0].strip()
    except Exception:
        fs_version = None

    # Get the version as described by git, if present.
    version = get_git_version()
    if version is None:
        version = fs_version

    if version is None:
        raise ValueError("Cannot find the version number!")

    if version != fs_version:
        f = open("RELEASE-VERSION", "w")
        f.write("{}\n".format(version))

    return version


setup(
    version=get_version(),
    packages=[""],
    package_dir={"": "slang"},
    cmake_source_dir="slang",
    cmake_install_dir="slang",
    cmake_install_target="pyslang-install-pylib",
    cmake_args=[
        "-DSLANG_INCLUDE_TESTS=OFF",
        "-DSLANG_INCLUDE_TOOLS=OFF",
        "-DSLANG_INCLUDE_PYLIB=ON",
    ],
    extras_require={"test": ["pytest"]},
)



#!/usr/bin/env python
# This script deletes the given directory and everything within it.
# cmake makes this difficult so here we are.
import shutil
import sys


def main():
    shutil.rmtree(sys.argv[1], ignore_errors=True)


if __name__ == "__main__":
    main()
